/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include "dixie.h"

#include <quipu/attrvalue.h>
#include <quipu/ds_error.h>
#include <quipu/commonarg.h>
#include <quipu/read.h>
#include <quipu/dua.h>

#define NULL	0

#define READ_END(x)	{ as_free(tempas); \
			 dn_free(thedn); \
			return(x); }

extern int	debug;
extern int	bound;
extern int	stream_connected;
extern short	timelimit, sizelimit;

/* args for the read call */
struct ds_read_arg	read_arg = { default_common_args };
struct ds_read_result	read_result;
struct DSError		read_error;

readop(bp, rp, rplenp, include_fileattrs, dontderef )
char	*bp, *rp;
int	*rplenp;
char	include_fileattrs;
char	dontderef;
{
	DN		thedn = NULLDN;
	Attr_Sequence	as_typelist(), tempas = NULLATTR;
	int		flag;
	int		i;
	int		rpsize;
	int		limit;
	PS		ps;
	int		len;

	read_arg.rda_common.ca_servicecontrol.svc_timelimit = timelimit;
	read_arg.rda_common.ca_servicecontrol.svc_sizelimit = sizelimit;
	if ( dontderef )
		read_arg.rda_common.ca_servicecontrol.svc_options |=
		    SVC_OPT_DONTDEREFERENCEALIAS;
	else
		read_arg.rda_common.ca_servicecontrol.svc_options &=
		    ~SVC_OPT_DONTDEREFERENCEALIAS;
	*rp = '\0';
	rpsize = *rplenp;
	*rplenp = 0;
	if ( debug & 2 ) printf("Read (%s)\n", bp);

	/* 
	 * First convert the dn given into a DN structure
	 */

	if ( *bp == '@') bp++;
	if ( (thedn = str2dn(bp)) == NULLDN ) {
		if ( debug > 1 ) printf("readops: str2dn (%s) failed\n",bp);
		READ_END(DXI_BADDN);
	}
	read_arg.rda_object = thedn;

	/*
	 * Next, step through all the attributes and construct
	 * the sequence of attributes to return from the read.  If no
	 * attributes are given, return all attributes.
	 */

	bp += strlen(bp) + 1;	/* skip the dn */
	read_arg.rda_eis.eis_select = as_typelist(bp, &flag);
	read_arg.rda_eis.eis_allattributes = flag;
	if ( read_arg.rda_eis.eis_select == NULLATTR && flag == 0 )
		READ_END(DXI_BAD_ATTRVAL);
	read_arg.rda_eis.eis_infotypes = EIS_ATTRIBUTESANDVALUES;

	read_result.rdr_entry.ent_attr = NULLATTR;
	read_result.rdr_entry.ent_dn = NULLDN;

	/* do the actual read */
	limit = 0;
	for ( i = 0; i < 2; i++ ) {
		if ( ds_read(&read_arg, &read_error, &read_result) == DS_OK )
			break;
		if ( debug & 2 ) printf("ds_read failed\n");

		if ( read_error.dse_type == DSE_REFERRAL )
			if ( chase_referral( &read_error ) != OK )
				break;
			else
				continue;

		if ( read_error.dse_type != DSE_REMOTEERROR
		    || stream_connected )
			break;

		(void) nullbind();

		if ( bound == 0 )
			break;
	}

	if ( debug & 2 )
		printf("read successful - checking for errors\n");

	switch ( read_error.dse_type ) {
	case DSE_REMOTEERROR:
		if ( debug & 2 ) {
			printf("\tservice error\n");
			error( &read_error );
		}
		READ_END(DXI_DSA_DOWN);
	case DSE_REFERRAL:
		if ( debug & 2 ) {
			printf("\treferral error\n");
			error( &read_error );
		}
		READ_END(DXI_REFER_ERROR);
	case DSE_NAMEERROR:
		if ( debug & 2 ) {
			printf("\tname error\n");
			error( &read_error );
		}
		READ_END(DXI_NAME_ERROR);
	case DSE_SERVICEERROR:
		if ( debug & 2 ) {
			printf("\tservice error\n");
			error( &read_error );
		}
		if (read_error.dse_un.dse_un_service.DSE_sv_problem
		    != DSE_SV_TIMELIMITEXCEEDED &&
		    read_error.dse_un.dse_un_service.DSE_sv_problem
		    != DSE_SV_ADMINLIMITEXCEEDED)
			READ_END(DXI_ERROR)
		else
			limit = 1;
		break;
	case DSE_ATTRIBUTEERROR:
		if ( debug & 2 ) {
			printf("\tattribute error\n");
			error( &read_error );
		}
		READ_END(DXI_NEX_ATTR);
	case DSE_NOERROR:
		if ( debug & 2 ) printf("\tno error\n");
		break;
	default:
		if ( debug & 2 ) {
			printf("\tsome unknown error (%d)\n",
			    read_error.dse_type);
			error( &read_error );
		}
		READ_END(DXI_ERROR);
	}
	dn_free(thedn);

	tempas = read_result.rdr_entry.ent_attr;
	thedn = read_result.rdr_entry.ent_dn;

	if ((ps = ps_alloc(str_open)) == NULLPS) {
		if (debug) printf("ps_alloc failed\n");
		clean_exit(1);
	}
	if (str_setup(ps, rp, rpsize, 1) == NOTOK) {
		if (debug) printf("str_setup failed\n");
		clean_exit(1);
	}

	if ( debug & 2 ) {
		printf("got a result!\n");
		fflush(stdout);
	}

	len = dn_sprint(ps, thedn);
	*rplenp = len;
	rpsize -= len;
	ps_print(ps, "\002");
	rp += len + 1;
	if ( tempas ) {
		len = as_sprint(ps, rp, tempas, rpsize,
		    include_fileattrs );
		*rplenp += len + 1;
	}

	ps_free(ps);

	READ_END(limit ? DXI_LIMIT : DXI_SUCCESS);
}
