/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include "dixie.h"
#include <quipu/attrvalue.h>
#include <quipu/ds_error.h>
#include <quipu/commonarg.h>
#include <quipu/ds_search.h>
#include <quipu/list.h>
#include <quipu/dua.h>

#define NULL	0
#define GOT_FILTER 666

#define SEARCH_END(x)	{\
		dn_free(thedn); \
		filter_free(search_arg.sra_filter); \
		as_free(search_arg.sra_eis.eis_select); \
		return(x); \
}

extern int	debug;
extern int	bound;
extern int	stream_connected;
extern short	timelimit, sizelimit;

/* args for the search call */
static struct ds_search_arg	search_arg = { default_common_args };
static struct ds_search_result	search_result;
static struct DSError		search_error;

searchop(bp, rp, type, rplen, header, dontderef, dontsearchaliases)
char	*bp, *rp;
int	type;
int	rplen;
char	*header;
char	dontderef;
char	dontsearchaliases;
{
	int		wild, flag, timeproblem, sizeproblem;
	int		rc;
	short		matches, netshort;
	DN		thedn = NULLDN;
	AttributeType	searchtype = NULLAttrT;
	AttributeValue	searchvalue = NULLAttrV;
	AV_Sequence	avs_comp_new();
	Attr_Sequence	as, as_typelist(), as_combine();
	EntryInfo	*matchlist;
	int		i, version;
	Filter		get_filter();

	search_arg.sra_common.ca_servicecontrol.svc_timelimit = timelimit;
	search_arg.sra_common.ca_servicecontrol.svc_sizelimit = sizelimit;
	if ( dontderef )
		search_arg.sra_common.ca_servicecontrol.svc_options |=
		    SVC_OPT_DONTDEREFERENCEALIAS;
	else
		search_arg.sra_common.ca_servicecontrol.svc_options &=
		    ~SVC_OPT_DONTDEREFERENCEALIAS;
	version = header[DXI_H_VERSION];
	search_arg.sra_eis.eis_select = NULLATTR;
	search_arg.sra_filter = NULLFILTER;
	search_arg.sra_searchaliases = !dontsearchaliases;
	search_arg.sra_eis.eis_infotypes = EIS_ATTRIBUTESANDVALUES;
	search_arg.sra_baseobject = NULLDN;
	search_result.srr_un.srr_unit = 0;

	matches = 0;
	if ( version == 0 )
		rp[0] = matches;
	else if ( version == 1 ) {
		netshort = htons(matches);
		(void) bcopy( (char *) &netshort, rp, sizeof(short) );
	}
	rp[2] = '\0';

	if ( *bp == '@' ) bp++;

	/*
	 * Now we go about the odious task of construcing the search
	 * arguments.  First the distinguished name at which to start
	 * searching, then the attribute type to search on, then the value 
	 * to search on, then attributes to return for matches.
	 */

	/* distinguished name */
	if ( *bp == '\0' )
		thedn = NULLDN;
	else if ( (thedn = str2dn(bp)) == NULLDN ) {
		if ( debug & 2 )
			printf("searchop: base str2dn of (%s) failed\n", bp);
		SEARCH_END(DXI_BADDN);
	}
	search_arg.sra_baseobject = thedn;

	bp += strlen(bp) + 1;

	/* search attribute type */
	if ( index( bp, '=' ) == 0 ) {
		if ( debug & 8 ) printf("old filter\n");
		if ( (searchtype = str2AttrT(bp)) == NULLAttrT ) {
			if ( debug & 2 ) 
				printf("searchop: (%s): No such attribute\n",
				    bp);
			SEARCH_END(DXI_ERROR);
		}

		search_arg.sra_filter = filter_alloc();
		if ( search_arg.sra_filter == NULLFILTER ) {
			perror("filter malloc");
			clean_exit(1);
		}
		search_arg.sra_filter->flt_type = FILTER_ITEM;
		search_arg.sra_filter->flt_next = NULLFILTER;

		/* go on to search attribute value */
		bp += strlen(bp) + 1;

		/* check for simple wildcard filter */
		if ( *bp == '*' && (type == DXI_SEARCH || type == DXI_SEARCH1
		    || type == DXI_SEARCH_EXACT)) {
			wild = 1;
			if ( debug ) printf("wild card filter\n");
		} else {
			wild = 0;
			if ( debug ) printf("regular filter\n");

			if ( (searchvalue = str2AttrV(bp,
			    searchtype->oa_syntax)) == NULLAttrV ) {
				if ( debug )
					printf("(%s): Bad attribute value\n",
					    bp);
				SEARCH_END(DXI_BAD_ATTRVAL);
			}
		}
	} else {
		if ( debug & 8 ) printf("new filter\n");
		if ( (search_arg.sra_filter = get_filter( bp ))
		    == NULLFILTER )
			SEARCH_END( DXI_BAD_FILTER );
		type = GOT_FILTER;
		wild = 0;
	}
	bp += strlen(bp) + 1;

	/* attribute types to return for matches */
	as = as_typelist(bp, &flag);
/*
	if ( (as = as_combine( as, "objectClass=alias" )) == NULLATTR ) {
		if ( debug & 2 ) printf( "as_combine failed\n" );
		SEARCH_END( DXI_ERROR );
	}
*/
	/* always return type searched on... */
	if ( searchtype != NULLAttrT ) {
		if ( (as = as_merge( as, as_comp_new( searchtype, NULLAV, 
		    NULLACL_INFO ) )) == NULLATTR ) {
			if ( debug & 2 ) printf( "as_merge failed\n" );
			SEARCH_END( DXI_ERROR );
		}
	}
	search_arg.sra_eis.eis_select = as;
	search_arg.sra_eis.eis_allattributes = flag;

	/* 
	 * select the type and extent of the search
	 */

	if ( header[DXI_H_SRCH_SCOPE] != 0 )
		search_arg.sra_subset = header[DXI_H_SRCH_SCOPE] - 1;
	else
		search_arg.sra_subset = SRA_WHOLESUBTREE;

	switch (type) {
	/* subtree approx with filtering */
	case DXI_SEARCH:
		if (debug & 8)
			printf("DXI_SEARCH\n");
		search_arg.sra_filter->flt_type = FILTER_ITEM;
		if ( wild == 1 ) {
			search_arg.sra_filter->FUITEM.fi_type =
			    FILTERITEM_PRESENT;
			search_arg.sra_filter->FUITEM.UNTYPE = searchtype;
		} else {
			search_arg.sra_filter->FUITEM.fi_type =
			    FILTERITEM_APPROX;
			search_arg.sra_filter->FUITEM.UNAVA.ava_type =
			    searchtype;
			search_arg.sra_filter->FUITEM.UNAVA.ava_value =
			    searchvalue;
		}
		break;

	/* one-level exact */
	case DXI_SEARCH1:
		if (debug & 8)
			printf("DXI_SEARCH1\n");
		search_arg.sra_filter->flt_type = FILTER_ITEM;
		if ( wild == 1 ) {
			search_arg.sra_filter->FUITEM.fi_type =
			    FILTERITEM_PRESENT;
			search_arg.sra_filter->FUITEM.UNTYPE = searchtype;
		} else {
			search_arg.sra_filter->FUITEM.fi_type =
			    FILTERITEM_EQUALITY;
			search_arg.sra_filter->FUITEM.UNAVA.ava_type =
			    searchtype;
			search_arg.sra_filter->FUITEM.UNAVA.ava_value =
			    searchvalue;
		}
		search_arg.sra_subset = SRA_ONELEVEL;
		break;

	/* subtree approx w/o filtering */
	case DXI_SEARCH_APPROX:
		if (debug & 8)
			printf("DXI_SEARCH_APPROX\n");
		search_arg.sra_filter->flt_type = FILTER_ITEM;
		search_arg.sra_filter->FUITEM.fi_type = FILTERITEM_APPROX;
		search_arg.sra_filter->FUITEM.UNAVA.ava_type = searchtype;
		search_arg.sra_filter->FUITEM.UNAVA.ava_value = searchvalue;
		break;

	/* subtree exact or wild w/o filtering */
	case DXI_SEARCH_EXACT:
		if (debug & 8)
			printf("DXI_SEARCH_EXACT\n");
		search_arg.sra_filter->flt_type = FILTER_ITEM;
		if ( wild == 1 ) {
			search_arg.sra_filter->FUITEM.fi_type =
			    FILTERITEM_PRESENT;
			search_arg.sra_filter->FUITEM.UNTYPE = searchtype;
		} else {
			search_arg.sra_filter->FUITEM.UNAVA.ava_type =
			    searchtype;
			search_arg.sra_filter->FUITEM.UNAVA.ava_value =
			    searchvalue;
			search_arg.sra_filter->FUITEM.fi_type =
			    FILTERITEM_EQUALITY;
		}
		break;

	case DXI_SEARCH_ISUB:
		if (debug & 8)
			printf("DXI_SEARCH_ISUB\n");
		search_arg.sra_filter->flt_type = FILTER_ITEM;
		search_arg.sra_filter->flt_un.flt_un_item.fi_type =
		    FILTERITEM_SUBSTRINGS;
		search_arg.sra_filter->FUITEM.UNSUB.fi_sub_type =
		    AttrT_cpy(searchtype);
		search_arg.sra_filter->FUITEM.UNSUB.fi_sub_initial =
		    avs_comp_new(AttrV_cpy(searchvalue));
		search_arg.sra_filter->FUITEM.UNSUB.fi_sub_any = NULLAV;
		search_arg.sra_filter->FUITEM.UNSUB.fi_sub_final = NULLAV;
		break;

	case DXI_SEARCH_FSUB:
		if (debug & 8)
			printf("DXI_SEARCH_FSUB\n");
		search_arg.sra_filter->flt_type = FILTER_ITEM;
		search_arg.sra_filter->flt_un.flt_un_item.fi_type =
		    FILTERITEM_SUBSTRINGS;
		search_arg.sra_filter->FUITEM.UNSUB.fi_sub_type =
		    AttrT_cpy(searchtype);
		search_arg.sra_filter->FUITEM.UNSUB.fi_sub_initial = NULLAV;
		search_arg.sra_filter->FUITEM.UNSUB.fi_sub_any = NULLAV;
		search_arg.sra_filter->FUITEM.UNSUB.fi_sub_final =
		    avs_comp_new(AttrV_cpy(searchvalue));
		break;

	case DXI_SEARCH_IFSUB:
		if (debug & 8)
			printf("DXI_SEARCH_IFSUB\n");
		search_arg.sra_filter->flt_type = FILTER_OR;
		search_arg.sra_filter->FUFILT = filter_alloc();
		if (search_arg.sra_filter->FUFILT == NULLFILTER) {
			if (debug) perror("filter malloc");
			SEARCH_END(DXI_ERROR);
		}
		search_arg.sra_filter->FUFILT->flt_type = FILTER_ITEM;
		search_arg.sra_filter->FUFILT->FUITEM.fi_type =
		    FILTERITEM_SUBSTRINGS;
		search_arg.sra_filter->FUFILT->FUITEM.UNSUB.fi_sub_type = 
		    AttrT_cpy(searchtype);
		search_arg.sra_filter->FUFILT->FUITEM.UNSUB.fi_sub_initial = 
		    avs_comp_new(AttrV_cpy(searchvalue));
		search_arg.sra_filter->FUFILT->FUITEM.UNSUB.fi_sub_any = NULLAV;
		search_arg.sra_filter->FUFILT->FUITEM.UNSUB.fi_sub_final = 
		    NULLAV;
		search_arg.sra_filter->FUFILT->flt_next = filter_alloc();
		if (search_arg.sra_filter->FUFILT->flt_next == NULLFILTER) {
			if (debug) perror("filter malloc");
			SEARCH_END(DXI_ERROR);
		}
		search_arg.sra_filter->FUFILT->flt_next->flt_next = NULLFILTER;
		search_arg.sra_filter->FUFILT->flt_next->flt_type = FILTER_ITEM;
		search_arg.sra_filter->FUFILT->flt_next->FUITEM.fi_type =
		    FILTERITEM_SUBSTRINGS;
		search_arg.sra_filter->FUFILT->flt_next->FUITEM.UNSUB.fi_sub_type = AttrT_cpy(searchtype);
		search_arg.sra_filter->FUFILT->flt_next->FUITEM.UNSUB.fi_sub_initial = NULLAV;
		search_arg.sra_filter->FUFILT->flt_next->FUITEM.UNSUB.fi_sub_any = NULLAV;
		search_arg.sra_filter->FUFILT->flt_next->FUITEM.UNSUB.fi_sub_final = avs_comp_new(AttrV_cpy(searchvalue));
		break;

	case GOT_FILTER:
		break;

	default:
		if (debug) printf("unknown search type (%d)\n", type);
		SEARCH_END(DXI_ERROR);
		break;
	}

	/*
	 * Now do the actual search.
	 */

	timeproblem = sizeproblem = 0;
	if ( debug & 2 ) printf("searching...\n");

	for ( i = 0; i < 2; i++ ) {
		if ( ds_search( &search_arg, &search_error, &search_result )
		    == DS_OK )
			break;
		if ( debug & 2 ) printf("ds_search failed\n");

		if ( search_error.dse_type == DSE_REFERRAL )
			if ( chase_referral( &search_error ) != OK )
				break;
			else
				continue;

		if ( search_error.dse_type != DSE_REMOTEERROR
		    || stream_connected )
			break;

		(void) nullbind();

		if ( bound == 0 )
			break;
	}

	if ( debug & 2 )
		printf("search successful - checking for errors\n");

	switch ( search_error.dse_type ) {
	case DSE_REMOTEERROR:
		if ( debug & 2 ) {
			printf("\tservice error\n");
			error( &search_error );
		}
		SEARCH_END(DXI_DSA_DOWN);
	case DSE_REFERRAL:
		if ( debug & 2 ) {
			printf("\treferral error\n");
			error( &search_error );
		}
		SEARCH_END(DXI_REFER_ERROR);
	case DSE_NAMEERROR:
		if ( debug & 2 ) {
			printf("\tname error\n");
			error( &search_error );
		}
		SEARCH_END(DXI_NAME_ERROR);
	case DSE_SERVICEERROR:
		if ( debug & 2 ) {
			printf("\tservice error\n");
			error( &search_error );
		}
		if ( search_error.dse_un.dse_un_service.DSE_sv_problem
		    == DSE_SV_TIMELIMITEXCEEDED )
			timeproblem = 1;
		else if ( search_error.dse_un.dse_un_service.DSE_sv_problem
		    == DSE_SV_ADMINLIMITEXCEEDED)
			sizeproblem = 1;
		else
			SEARCH_END(DXI_ERROR)
		break;
	case DSE_NOERROR:
		if ( debug & 2 ) printf("\tno error\n");
		break;
	case DSE_SECURITYERROR:
		if ( debug & 2 ) printf("\tsecurity error\n");
		SEARCH_END(DXI_REFUSED);
		break;
	default:
		if ( debug & 2 ) {
			printf("\tsome unknown error (%d)\n",
			    search_error.dse_type);
			error( &search_error );
		}
		SEARCH_END(DXI_ERROR);
	}

	/* correlate results if necessary */
	if ( !search_result.srr_correlated ) {
		if ( debug & 2 ) printf("correlating results\n");
		correlate_search_results(&search_result);
	}

	/* some limit was exceeded */
	if ( search_result.CSR_limitproblem != LSR_NOLIMITPROBLEM ) {
		if ( debug ) printf("limit problem\n");
	}
	timeproblem |= (search_result.CSR_limitproblem
	    == LSR_TIMELIMITEXCEEDED);
	sizeproblem |= (search_result.CSR_limitproblem == LSR_SIZELIMITEXCEEDED
	    || search_result.CSR_limitproblem == LSR_ADMINSIZEEXCEEDED);

	/* nothing matched */
	if ( search_result.CSR_entries == NULLENTRYINFO ) {
		if ( debug & 2 ) printf("no entries returned\n");
		if ( timeproblem )
			rc = DXI_TIMELIMIT;
		else if ( sizeproblem )
			rc = DXI_SIZELIMIT;
		else
			rc = DXI_NOTFOUND;
		SEARCH_END(rc);
	}

	matchlist = NULLENTRYINFO;
	/* do filtering */
	if ( wild == 0 && type == DXI_SEARCH ) {
		matches = count_matches( search_result.CSR_entries, searchtype,
		    searchvalue, &matchlist );
		if ( debug & 2 ) 
			printf("found (%d) exact matches\n", matches);
	}

	if ( matchlist == NULLENTRYINFO ) {
		if ( debug & 2 ) printf( "printing entries\n" );
		matches = entry_sprint(version == 0 ? rp + 1 : rp + 2,
		    search_result.CSR_entries, rplen);
		if ( debug & 2 ) printf( "found (%d) matches\n", matches );
	} else {
		if ( debug & 2 ) printf( "printing matchlist entries\n" );
		matches = entry_sprint(version == 0 ? rp + 1 : rp + 2,
		    matchlist, rplen);
		if ( debug & 2 ) printf( "found (%d) matchlist matches\n",
			matches );
	}

	if ( version == 0 )
		rp[0] = matches;
	else if ( version == 1 ) {
		netshort = htons(matches);
		(void) bcopy( (char *) &netshort, rp, sizeof(short) );
	}

	dn_free( search_result.CSR_object );
	crefs_free( search_result.CSR_cr );
	entryinfo_free( search_result.CSR_entries, 0 );

	if ( timeproblem )
		rc = DXI_TIMELIMIT;
	else if ( sizeproblem )
		rc = DXI_SIZELIMIT;
	else
		rc = DXI_SUCCESS;
	SEARCH_END(rc)
}
